#pragma once

#include <assert.h>

#include <iosfwd>
#include <memory>
#include <thread>
#include <condition_variable>
#include <mutex>
#include <vector>

#include "memory_stream.hpp"
// async

// InputStream must be buffered to be useful, otherwise it doesn't make sense.
// mutex is not movable so we need an implemention that will be allocated
// and wrapped inside a unique_ptr.
template<typename T>
class AsyncInputStream_impl {
    typedef std::thread thread;
    typedef std::mutex mutex;
    typedef std::condition_variable condition_variable;
public:
    AsyncInputStream_impl               (AsyncInputStream_impl&&)       = delete;
    AsyncInputStream_impl &operator=    (AsyncInputStream_impl&& other) = delete;
    AsyncInputStream_impl               (const AsyncInputStream_impl&)  = delete;
    AsyncInputStream_impl &operator=    (const AsyncInputStream_impl&)  = delete;

    AsyncInputStream_impl() {
        mBlockSize  = 2048;
        mMaxBlocks  = 16;
        mRunning    = false;
        mFinished   = false;

        mBlocks.reserve(mMaxBlocks);
    }

    AsyncInputStream_impl(T &&stream, std::streamsize blockSize = 2048,
        std::streamsize maxBlocks = 16) : mStream(std::move(stream)) {
        set_block_size(blockSize, maxBlocks);
        mRunning    = false;
        mFinished   = false;

        mBlocks.reserve(mMaxBlocks);
    }

    ~AsyncInputStream_impl() {
        quitThread();
    }

    void set_block_size(std::streamsize blockSize, std::streamsize maxBlocks) {
        mBlockSize  = blockSize;
        mMaxBlocks  = maxBlocks;
        mBlocks.reserve(mMaxBlocks);
    }

    std::streamsize read(void* data, std::streamsize size) {
        /* @todo: some people will not realize start_if_needed() or try to hack
                  by doing a read(nullptr, 0) as a higher level class might
                  abstract away this interface. Should this be allowed?
        */
        start_if_needed();
        if(size <= 0) {
            return 0;
        }
        ExistingMemoryBlock block(data, size);

        std::unique_lock<mutex> lock(mMutex);
        mCondition.wait(lock, [&] () -> bool {
            return !mRunning || !mBlocks.empty();
        });
        lock.unlock();
        std::streamsize totalRead = 0;
        while(true) {
            if(mBlocks.empty()){
                break;
            }
            assert(mBlocks[0].size() != mBlocks[0].tell());
            std::streamsize bRead = mBlocks[0].read(block.cursor(), block.capacity() - block.tell());
            if(mBlocks[0].size() == mBlocks[0].tell()) {
                lock.lock();
                mBlocks.erase(mBlocks.begin());
                mCondition.notify_all();
                lock.unlock();
            }
            if(bRead > 0) {
                block.seek(bRead, std::ios_base::cur);
                totalRead += bRead;
            }
            if(block.tell() == block.capacity()) {
                break;
            }
        }
        return totalRead;
    }

    void start_if_needed() {
        if(mRunning || mFinished) {
            return;
        }
        mRunning = true;
        mThread = thread([this]{ this->readerThread(); });
    }

    /* We need get/set to allow chaining for filter */
    void set_istream(T &&stream) {
        mStream = std::move(stream);
    }

    T &get_istream() {
        return mStream;
    }
private:
    /** @returns true if block is full */
    std::streamsize readIntoBlock(MemoryDuplexStream &block) {
        std::streamsize remaining = block.capacity() - block.tell();
        assert(remaining != 0);
        if(remaining == 0){
            return 0;
        }
        std::streamsize bRead = mStream.read(block.cursor(), remaining);
        if(bRead > 0){
            block.write(bRead);
        }
        return bRead;
    }
    void readerThread() {
        MemoryDuplexStream curBlock(mBlockSize);
        while(true) {
            int numBlocks = 0;
            {
                std::unique_lock<mutex> lock(mMutex);
                mCondition.wait(lock, [&] () -> bool {
                    return !mRunning || canReadMore();
                });
                if(!mRunning) {
                    if(curBlock.tell() > 0){
                        curBlock.seek(0);
                        mBlocks.push_back(std::move(curBlock));
                    }
                    break;
                }
                numBlocks = mBlocks.size();
            }
            std::streamsize bRead = readIntoBlock(curBlock);
            bool isFull = curBlock.tell() == curBlock.capacity();
            bool hasData = curBlock.tell() > 0;
            if(hasData && (isFull || numBlocks == 0 || bRead <= 0)) {
                curBlock.truncate(curBlock.tell());
                {
                    std::unique_lock<mutex> lock(mMutex);
                    curBlock.seek(0);
                    mBlocks.push_back(std::move(curBlock));
                    mCondition.notify_all();
                }
                curBlock.set_capacity(mBlockSize);
                curBlock.truncate(0);
            }
            if(bRead == 0) {
                std::unique_lock<mutex> lock(mMutex);
                mRunning = false;
                mFinished = true;
                mCondition.notify_all();
                break;
            }
        }

    }
    bool canReadMore() {
        return static_cast<std::streamsize>(mBlocks.size()) < mMaxBlocks;
    }

    void quitThread() {
        {
            std::unique_lock<mutex> lock(mMutex);
            mRunning = false;
            mCondition.notify_all();
        }
        if(mThread.joinable()){
            mThread.join();
        }
    }

    T                           mStream;
    thread                      mThread;
    mutex                       mMutex;
    condition_variable          mCondition;

    std::vector<MemoryDuplexStream> mBlocks;
    std::streamsize             mBlockSize;
    std::streamsize             mMaxBlocks;
    bool                        mRunning;
    bool                        mFinished;
};

struct defer_t{};
constexpr defer_t defer;

template<typename T>
class async_istream {
    typedef AsyncInputStream_impl<T> PrivateImp;
public:
    async_istream            (async_istream&&)        =default;
    async_istream &operator= (async_istream&&)        =default;
    async_istream            (const async_istream&)   =delete;
    async_istream &operator= (const async_istream)    =delete;

    /** You must explicitly call start_if_needed, or on first read() the read
        ahead thread will be spawned.
    */
    async_istream(std::streamsize blockSize = 2048,
        std::streamsize maxBlocks = 16) {
        mImpl.reset(new PrivateImp(blockSize, maxBlocks));
    }

    /** Emediatly starts thread reading the stream */
    async_istream(T &&stream, std::streamsize blockSize = 2048,
        std::streamsize maxBlocks = 16) {
        mImpl.reset(new PrivateImp(std::move(stream), blockSize, maxBlocks));
        start_if_needed();
    }

    std::streamsize read(void* data, std::streamsize size) {
        return mImpl->read(data, size);
    }

    /* get/set is needed for chaining */

    void set_istream(T &&stream) {
        mImpl->set_istream(std::move(stream));
    }

    T &get_istream() {
        return mImpl->getInputStream();
    }

    void start_if_needed() {
        mImpl->start_if_needed();
    }

private:
    std::unique_ptr<PrivateImp> mImpl;
};
