#pragma once

#include <iosfwd>
#include <cstdint>

/* satisfying everyone is hard. */

/*
    What do people need/want

    - automatically destroy memory (RAII)
    - ability to not own the memory and use specific location. Maybe for use
      in some memory mapped interprocess communication.
    - simple
    - std::vector but it's missing the following.
        - Can't specify specific memory address for buffer.
        - constant sizeof(...) size for different kinds of allocators.
        - abstract away allocator.

    All those are conflicting. Either we be like vector, or not own any data
    at all. Doing both is super complicated.

    after much thought I think the best way is to design the streams no not
    take any additional buffers. Buffers add much complexity. Doing it this way
    reduces the complexity. And a person can write their own buffer stream
    if they want to use a specific memory location. Unfortunatly that's alot
    of work. Maybe there will be a way to make it easier.

*/
/**
    The buffer expects data to be allocated with new uint8_t[]. It will
    call `delete[]` on the starting address of the data pointer passed in. If
    more complex destruction is needed than an allocator is to be attached to
    object that will own the buffer. That object will destroy it appropriately.
    The method leak() is provided to leak data. This is to be used with said
    deleter so that buffer does not delete the data pointer itself.
*/
#if 0
class buffer {
public:
    buffer(std::streamsize bufferSize) {
        mDataPtr    = new uint8_t[bufferSize];
        mDataPtrEnd = mDataPtr + bufferSize;
    }
    buffer(uint8_t *data, uint8_t *end) {
        mDataPtr = reinterpret_cast<uint8_t*>(data);
        mDataPtrEnd = end;
    }

    /** To make a copy we need an allocator defined. */
    buffer(const buffer&)=delete;
    buffer& operator=(const buffer&)=delete;

    buffer(buffer&& other) : mDeleter(std::move(other.mDeleter)) {
        mDataPtr            = other.mDataPtr;
        mDataPtrEnd         = other.mDataPtrEnd;
        other.mDataPtr      = nullptr;
        other.mDataPtrEnd   = nullptr;
    }
    /*
        This adds quite a bit complication to deleter. I'm not going to have it
        in just to avoid some complexities. We can live without it anyway.
    */
    buffer &operator=(buffer&& other)=delete;

    void leak() {
        mDataPtr    = nullptr;
        mDataPtrEnd = nullptr;
    }
    ~buffer() {
        if(mDataPtr) {
            delete [] mDataPtr;
            mDataPtr    = nullptr;
            mDataPtrEnd = nullptr;
        }
    }

    /*
        @returns true if other is exactly the same size and data is equal to
                 byte for byte to this.
    */
    bool operator==(const buffer &other) const;
    bool operator!=(const buffer &other) const { return !(*this == other); }

    template<typename T=uint8_t>
    T *get() { return reinterpret_cast<T*>(mDataPtr); }
    std::streamsize size() { return mDataPtrEnd - mDataPtr; }

    uint8_t *begin() const {
        return mDataPtr;
    }
    uint8_t *end() const {
        return mDataPtrEnd;
    }
private:
    uint8_t *mDataPtr;
    uint8_t *mDataPtrEnd;
};

#endif
