#pragma once

#include <memory>

template<typename T, typename Deleter=std::default_delete<T>>
class MaybeDeleter {
public:
    MaybeDeleter(MaybeDeleter&&)=default;
    MaybeDeleter(const MaybeDeleter&)=default;
    MaybeDeleter() {
        mOwns = false;
    }
    
    template<typename DeleterForward>
    MaybeDeleter(DeleterForward&& other) : mDeleter(std::forward<DeleterForward>(other)) {
        mOwns = true;
    }
    void operator()(T *object) {
        if(mOwns) {
            mDeleter(object);
        }
    }
private:
    bool mOwns;
    Deleter mDeleter;
};

// We need Data not be templated and be passable from one object to another
// very freely.

// not DataBase like SQL. Base as in Base class.
/* 
    I want this data class to be very self contained. Even the allocater
    used to allocate the base object too. Or same way of.
    
    This is a similar trick std::shared_ptr does to abstract away the delete.
*/
class DataBase {
public:
    virtual ~DataBase(){}
    
    template<typename T=uint8_t>
    T *get() { return reinterpret_cast<T*>(mDataPtr); }
    std::streamsize size() { return mSize; }
    
    uint8_t *begin() const {
        return mDataPtr;
    }
    uint8_t *end() const {
        return mDataPtr + mSize;
    }
    
    void copyData(const DataBase &other) {
        std::streamsize size = std::min(other.mSize, mSize);
        std::copy(other.begin(), other.begin()+size, mDataPtr);
    }
    
    /** 
        @returns true if other is exactly the same size and data is equal to
                 byte for byte to this.
    */
    bool operator==(const DataBase &other) const;
    bool operator!=(const DataBase &other) const { return !(*this == other); }
protected:
    uint8_t *mDataPtr;
    int mSize;
};

/* 
    Some times the data is allocated on the stack or somewhere else. In which
    case we don't want the stream to own the data and destroy it.
*/
template<typename T>
struct no_deleter {
    void operator()(T *object) {
        // do nothing.
    }
};

template<typename Deleter=std::default_delete<uint8_t[]>>
class Data : public DataBase {
public:
    template<typename D>
    Data(uint8_t *data, std::streamsize size, D&& deleter) 
        : mDeleter(std::forward<Deleter>(deleter)) {
        mDataPtr = data;
        mSize = size;
    }

    Data(uint8_t *data, std::streamsize size) {
        mDataPtr = reinterpret_cast<uint8_t*>(data);
        mSize = size;
    }
    
    /** To make a copy we need an allocator defined. */
    Data(const Data&)=delete;
    Data& operator=(const Data&)=delete;
    
    Data(Data&& other) : mDeleter(std::move(other.mDeleter)) {
        mDataPtr = other.mDataPtr;
        other.mDataPtr = nullptr;
    }
    /* 
        This adds quite a bit complication to deleter. I'm not going to have it
        in just to avoid some complexities. We can live without it anyway.
    */
    Data &operator=(Data&& other)=delete;
    
    virtual ~Data() {
        if(mDataPtr) {
            mDeleter(mDataPtr);
            mDataPtr = 0;
            mSize = 0;
        }
    }
    
private:
    Deleter mDeleter;
};

/*  
    We want this to transfer the data into the stream. The tricky part is
    the stream might have a different allocater than the original Data object.
    So the stream will create it's own copy of the data object using it's own
    allocator then simply move the data object in.
*/
template<typename Allocator, typename Deleter>
Data<Deleter> *takeOwnership(Data<Deleter> &&data, Allocator &allocator) {
    Data<Deleter> *newData = allocator.allocate(1);
    new (newData) Data<Deleter>(std::move(data));
    return newData;
}

Data<> loadFile(const std::string &fileName);
