#pragma once

// We just need to forward declare FILE* which is not possible. So we must
// include this header
#include <cstdio>

#include <iosfwd>
#include <ios>

/* the libstd++ source files have __file_type which abstracts away OS specific
   API. This bellow is a poor reimplementation ontop of C FILE. This class
   is meant for easier adapting into libstd++. It appears __file_type is
   not buffured. This class will behave so.

   This will be OS specific so it's core implementation is not in the header.
   But it could be in the header if done right. Looking at existing libstd++
   they haven't put the core implementation in the header so we will do
   similar.
*/
class FileType {
public:
    typedef FILE *native_handle_type;
    FileType();
    ~FileType();

    FileType(FileType&& other) {
        (*this) = std::move(other);
    }
    FileType &operator=(FileType&& other) {
        mFile = other.mFile;
        other.mFile = nullptr;
        return *this;
    }
    FileType(const FileType&)=delete;
    FileType &operator= (const FileType&)=delete;
    explicit FileType(const std::string &fileName, std::ios_base::openmode openMode) {
        open(fileName, openMode);
    }
    bool open(const std::string &fileName, std::ios_base::openmode openMode);

    bool setBufferSize(std::streamsize size);

    /* equivalent of xsgetn */
    std::streamsize read(void *data, std::streamsize size);
    /* equivalent of xsputn */
    std::streamsize write(const void *data, std::streamsize size);

    std::streamsize write_2(const void* data1, std::streamsize size1,
        const void *data2, std::streamsize size2);
    std::streamoff seek(std::streamoff offset, std::ios_base::seekdir dir = std::ios_base::beg);

    std::streamoff tell() const;
    void close();
    bool isOpen() const;
private:
    native_handle_type mFile;
};


/* Just an example of input stream & output stream. Strictly binary stremas */

class file_istream {
public:
    file_istream() {
        // nothing to do
    }
    // truncate & append doesn't make sense here. ate could make sense
    // I think it's better to leave the parameter out and user can explicitly
    // just call seek to go to end
    explicit file_istream(const std::string &fileName){
        open(fileName);
    }
    bool open(const std::string &fileName) {
        return mFile.open(fileName, std::ios_base::in | std::ios_base::binary);
    }

    std::streamsize read(void *data, std::streamsize nbytes) {
        return mFile.read(data, nbytes);
    }
    std::streamoff seek(std::streamoff offset, std::ios_base::seekdir dir = std::ios_base::beg) {
        return mFile.seek(offset, dir);
    }
    std::streamoff tell() const { return mFile.tell(); }
    bool is_open() const { return mFile.isOpen(); }
private:
    FileType mFile;
};

class file_ostream {
public:
    file_ostream() {
        // nothing to do
    }
    explicit file_ostream(std::string fileName, std::ios_base::openmode openMode) {
        open(fileName, openMode | std::ios_base::binary);
    }
    bool open(const std::string &fileName, std::ios_base::openmode openMode) {
        return mFile.open(fileName, openMode | std::ios_base::out | std::ios_base::binary);
    }
    std::streamsize write(const void *data, std::streamsize nbytes) {
        return mFile.write(data, nbytes);
    }
    std::streamoff tell() const { return mFile.tell(); }
    bool is_open() const { return mFile.isOpen(); }
private:
    FileType mFile;
};

class file_iostream {
public:
    file_iostream(){}
    explicit file_iostream(const std::string &fileName, std::ios_base::openmode openMode) {
        open(fileName, openMode);
    }
    bool open(const std::string &fileName, std::ios_base::openmode openMode) {
        return mFile.open(fileName, openMode | std::ios_base::binary);
    }
    std::streamsize read(void *data, std::streamsize nbytes) {
        return mFile.read(data, nbytes);
    }
    std::streamsize write(const void *data, std::streamsize nbytes) {
        return mFile.write(data, nbytes);
    }
    std::streamoff seek(std::streamoff offset, std::ios_base::seekdir dir = std::ios_base::beg) {
        return mFile.seek(offset, dir);
    }
    std::streamoff tell() const { return mFile.tell(); }
    bool is_open() const { return mFile.isOpen(); }

private:
    FileType mFile;

};
