#pragma once


#include <assert.h>
// streamsize
#include <iosfwd>
#include <ios>
#include <climits>

#include "data.hpp"

template<typename T>
class LineIterator {
public:
    typedef char char_type;
    LineIterator() {
    }
    LineIterator(T &stream) : mStream(&stream) { assert(mStream != nullptr); }
    void readNextLine() {
        char_type c;
        mCurLine.clear();
        if(mStream == nullptr) {
            return;
        }
        do {
            std::streamsize bRead = mStream->read(&c, sizeof(char_type));
            if(bRead <= 0){
                mStream = nullptr;
                break;
            }
            if(c == '\r'){
                mHasCarrage = true;
                break;
            }
            if(c == '\n' && !mHasCarrage){
                break;
            }
            mCurLine += c;
        } while(true);
        ++mLineNumber;
    }

    const std::string &operator*() {
        if(mNeedsRead){
            readNextLine();
            mNeedsRead = false;
        }
        return mCurLine;
    }
    LineIterator &operator++() {
        readNextLine();
        return *this;
    }

    bool operator==(const LineIterator<T> &other) const {
        if(this == &other){
            return true;
        }
        if(other.mStream != mStream) {
            return false;
        }
        if(other.mLineNumber == mLineNumber) {
            return true;
        }
        if(other.mStream == nullptr && mCurLine.empty()) {
            return true;
        }
        return false;
    }

    bool operator!=(const LineIterator<T> &other) const {
        return !((*this) == other);
    }
private:
    // iterator does not own the stream
    // can be nullptr in the case of when it's the end
    T *mStream = nullptr;
    std::string mCurLine;
    bool mNeedsRead = true;
    bool mHasCarrage = false;
    int mLineNumber = 0;
};

template<typename T>
class LineReader {
public:
    typedef LineIterator<T> iterator;
    LineReader              (LineReader&&)      =default;
    LineReader &operator=   (LineReader&&)      =default;
    LineReader              (const LineReader&) =delete;
    LineReader &operator=   (const LineReader&) =delete;

    // take ownership
    LineReader(T &&stream) : mStream(std::move(stream)) {
    }
    iterator begin() {
        // you can only begin once. Any more and it is an error
        assert(mDidBegin == false);
        mDidBegin = true;
        return iterator(mStream);
    }
    iterator end() {
        return iterator();
    }

    /* get/set is needed for chaining */

    void setInputStream(T &&stream) {
        mStream = std::move(stream);
    }

    T &getInputStream() {
        return mStream;
    }
private:
    T mStream;
    bool mDidBegin = false;
};

// class requires <>. As function it can be autodetected and skip <>.
template<typename T>
LineReader<T> lineReader(T &&stream) {
    return LineReader<T>(std::move(stream));
}




// source: http://stackoverflow.com/a/4956493/884893
// some processors support byteswapping instructions set.
template <typename T>
T swap_endian(T u)
{
    static_assert (CHAR_BIT == 8, "CHAR_BIT != 8");

    union
    {
        T u;
        unsigned char u8[sizeof(T)];
    } source, dest;

    source.u = u;

    for (size_t k = 0; k < sizeof(T); k++)
        dest.u8[k] = source.u8[sizeof(T) - k - 1];

    return dest.u;
}


/*
    There is no constexpr way to do this because some processors can change
    endianess at runtime.
*/
// source: http://stackoverflow.com/a/4240014/884893
// modified from source
inline bool is_little_endian() {
    short int number = 0x1;
    static_assert(sizeof(number) >= 2, "not enough bits in short to determin endians");
    uint8_t *numPtr = (uint8_t*)&number;
    return (numPtr[0] == 1);
}

inline bool is_big_endian() {
    return !is_little_endian();
}

template<typename T>
T to_big_endian(T value) {
    return is_big_endian()? value : swap_endian(value);
}
template<typename T>
T to_little_endian(T value) {
    return is_little_endian()? value : swap_endian(value);
}

template<typename InputStream>
std::streamsize skip(InputStream &stream, std::streamsize skipAmount,
    uint8_t *gargbageStart, uint8_t *gargbageEnd) {
    std::streamsize totalSkipped = 0;

    std::streamsize maxSkipIteration = gargbageEnd - gargbageStart;
    while(totalSkipped < skipAmount) {
        std::streamsize skip = skipAmount - totalSkipped;
        skip = std::min(skip, maxSkipIteration);
        std::streamsize didSkip = stream.read(gargbageStart, skip);
        if(didSkip <= 0) {
            break;
        }
        totalSkipped += didSkip;
    }
    return totalSkipped;
}

struct copy_config {
    copy_config() {
        buffer_begin    = nullptr;
        buffer_end      = nullptr;
    }
    copy_config(void *buffer_begin, std::streamsize size) {
        this->buffer_begin  = reinterpret_cast<uint8_t*>(buffer_begin);
        this->buffer_end    = this->buffer_begin + size;
    }
    copy_config(void *buffer_begin, void *buffer_end) {
        this->buffer_begin  = reinterpret_cast<uint8_t*>(buffer_begin);
        this->buffer_end    = reinterpret_cast<uint8_t*>(buffer_end);
    }
    uint8_t *buffer_begin;
    uint8_t *buffer_end;
};

/** Writes data until either a write operation returns 0 or less or written
    `size` amount of bytes.

    @param output       the output stream.
    @param begin_ptr    the starting of data to write.
    @param size         size in bytes to write.

    @returns total bytes written.
*/
template<typename Output>
std::streamsize write_fully(Output &output, const void *begin_ptr, std::streamsize size) {
    int iteration = 0;
    std::streamsize total_write = 0;

    const uint8_t *begin    = reinterpret_cast<const uint8_t*>(begin_ptr);
    const uint8_t *end      = begin + size;
    while(begin < end) {
        std::streamsize expected_write = end - begin;
        std::streamsize did_write = output.write(begin, expected_write);
        if(did_write <= 0) {
            break;
        }
        total_write += did_write;
        begin += did_write;
        ++iteration;
    }
    return total_write;
}

/** Tries to read the input a full amount of `size` bytes. Stops reading when
    a read operation returns 0 or negative.

    @param input        The input stream.
    @param begin_ptr    ptr to save the data into.
    @param size         size in bytes to read.

    @returns total bytes read.
*/

template<typename Input>
std::streamsize read_fully(Input &input, void *begin_ptr, std::streamsize size) {
    int iteration = 0;
    std::streamsize total_read = 0;

    uint8_t *begin    = reinterpret_cast<uint8_t*>(begin_ptr);
    uint8_t *end      = begin + size;
    while(begin < end) {
        std::streamsize expected_read = end - begin;
        std::streamsize did_read = input.read(begin, expected_read);
        if(did_read <= 0) {
            break;
        }
        total_read += did_read;
        begin += did_read;
        ++iteration;
    }
    return total_read;
}

struct copy_result {
    /** Total bytes read from input */
    std::streamsize total_read = 0;
    /** Total bytes written to output */
    std::streamsize total_write = 0;
    /** true if did_read == did_write and greator than 0 bytes copied */
    bool success = false;

    explicit operator bool() const {
        return success;
    }
};

// private internal class
namespace internal {
    class buffer_guard {
    public:
        ~buffer_guard() {
            if(buffer) {
                delete [] buffer;
            }
        }
        uint8_t *allocate(std::streamsize size) {
            buffer = new uint8_t[size];
            return buffer;
        }
        uint8_t *buffer = nullptr;
    };
}

/** Copy data from input to output stream.

    @code
        // use a default heap allocated buffer.
        copy_results results = copy(input, output);
        // if you only care about success/failure
        bool success = copy(input, output).success;

        // example with custom buffer
        std::vector<uint8_t> buffer(2048);
        copy_config config(buffer.data(), buffer.size());
        copy(input, output, config);
    @endcode
*/
template<typename Input, typename Output>
copy_result copy(Input &input, Output &output, copy_config config = copy_config()) {
    // incase an exception will be thrown
    internal::buffer_guard buffer;

    if(!config.buffer_begin) {
        constexpr std::streamsize size = 2048;
        config.buffer_begin = buffer.allocate(size);
        config.buffer_end   = config.buffer_begin + size;
    }

    std::streamsize size = config.buffer_end - config.buffer_begin;
    copy_result results;

    while(true) {
        std::streamsize did_read = input.read(config.buffer_begin, size);
        if(did_read <= 0) {
            break;
        }
        results.total_read += did_read;
        std::streamsize did_write = write_fully(output, config.buffer_begin, did_read);
        if(did_write > 0) {
            results.total_write += did_write;
        }
        if(did_write < did_read) {
            // @todo: what should we do? throw exception?
            break;
        }
    }

    if(results.total_read > 0) {
        results.success = results.total_read == results.total_write;
    }

    return results;
}
/*
// Example program

#include "iostream.hpp"
#include <iostream>

int main(){

    FileInputStream inputFile("testFile.txt");

    // allow iterating over text file yay :)
    for(const std::string &line : lineReader(inputFile)) {
        std::cout << line << '\n';
    }
    return 0;
}
*/
