#pragma once

#include <memory>

/*
    3 streams we need here

    - memory_helper_iostream
    - memory_iostream
    - heap_iostream
*/

template<typename pointer=uint8_t*>
class basic_memory_helper_iostream {
public:
    basic_memory_helper_iostream() : basic_memory_helper_iostream<pointer>((pointer)nullptr, (pointer)nullptr) {}

    /**
        @param start the first byte
        @param end 1 past the last byte in the sequence.
    */
    template<typename input_pointer=void*, typename input_pointer2=input_pointer>
    basic_memory_helper_iostream(input_pointer start, input_pointer end) {
        mStart          = reinterpret_cast<pointer>(start);
        mEnd            = reinterpret_cast<pointer>(end);
        mEndOfStorage   = reinterpret_cast<pointer>(end);
        mCursor         = mStart;
    }

    template<typename input_pointer=void*>
    basic_memory_helper_iostream(input_pointer start, std::streamsize size) {
        mStart          = reinterpret_cast<pointer>(start);
        mEnd            = mStart + size;
        mEndOfStorage   = mStart + size;
        mCursor         = mStart;
    }

    std::streamsize size() const {
        return mEnd - mStart;
    }

    bool eof() const {
        return mCursor == mEnd;
    }

    std::streamsize capacity() const {
        return mEndOfStorage - mStart;
    }

    std::streamsize write(const void *data, std::streamsize bSize) {
        pointer endWrite = mCursor + bSize;
        if(endWrite > mEndOfStorage) {
            endWrite = mEndOfStorage;
            bSize = mEndOfStorage - mCursor;
        }
        if(data != mCursor) {
            std::copy(
                reinterpret_cast<const uint8_t*>(data),
                reinterpret_cast<const uint8_t*>(data)+bSize, mCursor);
        }
        mCursor += bSize;
        if(endWrite > mEnd) {
            mEnd = endWrite;
        }
        return bSize;
    }

    std::streamsize write(std::streamsize size) {
        return write(mCursor, size);
    }

    std::streamsize read(void *data, std::streamsize bSize) {
        uint8_t *endRead = mCursor + bSize;
        if(endRead > mEnd) {
            endRead = mEnd;
            bSize = mEnd - mCursor;
        }
        if(bSize == 0) {
            return 0;
        }
        std::copy(mCursor, mCursor+bSize, reinterpret_cast<uint8_t*>(data));
        mCursor += bSize;
        return bSize;
    }
    std::streamsize tell() const {
        return mCursor - mStart;
    }

    std::streamoff seek(std::streamoff offset, std::ios_base::seekdir dir = std::ios_base::beg) {
        // @note fseek doesn't support 64bit on 32bit systems
        pointer relative;
        switch(dir) {
        default:
            // @todo error exception?
        case std::ios_base::beg: relative = mStart;     break;
        case std::ios_base::cur: relative = mCursor;    break;
        case std::ios_base::end: relative = mEnd;       break;
        }

        mCursor = relative + offset;
        if(mCursor > mEnd) {
            mCursor = mEnd;
        } else if (mCursor < mStart) {
            mCursor = mStart;
        }
        return tell();
    }

    void truncate(std::streamsize size) {
        mEnd = mStart + size;
        if(mEnd > mEndOfStorage) {
            mEnd = mEndOfStorage;
        }
        if(mCursor > mEnd) {
            mCursor = mEnd;
        }
    }

    pointer cursor() const { return mCursor; }
    pointer data() const { return mStart; }
private:
    pointer mStart; // the first byte
    pointer mEnd; // 1 past the last byte
    pointer mEndOfStorage; // 1 past the last byte.
    pointer mCursor; // next read will read this byte
};

typedef basic_memory_helper_iostream<uint8_t*> memory_helper_iostream;
typedef basic_memory_helper_iostream<const uint8_t*> const_memory_helper_iostream;
template<typename T, typename Allocator>
struct Deleter {
    void operator()(T *data) {

    }
private:
    std::streamsize mCount;
    Allocator mAllocator;
};


template<typename Allocator=std::allocator<uint8_t>>
class memory_iostream : public memory_helper_iostream {
public:
    template<typename Allocator2=Allocator>
    memory_iostream(std::streamsize size, const Allocator2 &allocator = Allocator())
        : mAllocator(allocator) {
        set_capacity(size);
    }
    memory_iostream(){}
    explicit memory_iostream(std::streamsize size) {
        set_capacity(size);
    }

    ~memory_iostream() {
        mAllocator.deallocate(data(), capacity());
    }

    memory_iostream &operator=(memory_iostream &&other) {
        if(other.mAllocator != mAllocator) {
            mAllocator = std::move(other.mAllocator);
        }
        *static_cast<memory_helper_iostream*>(this) = other;
        *static_cast<memory_helper_iostream*>(&other) = memory_helper_iostream();

        return *this;
    }
    memory_iostream(const memory_iostream&)             =default;
    memory_iostream(memory_iostream&& other) { *this = std::move(other); }


    // todo: templatize this to accept allocator
    bool set_capacity(std::streamsize newCapacity, bool copy = true) {
        if(capacity() == newCapacity) {
            return true;
        }
        std::streamsize offset = tell();
        std::streamsize originalSize = size();

        uint8_t *newData = mAllocator.allocate(newCapacity);
        if(data() != nullptr) {
            if(copy) {
                std::streamsize copySize = std::min(size(), newCapacity);
                std::copy(data(), data()+copySize, newData);
            }
            mAllocator.deallocate(data(), capacity());
        }

        *static_cast<memory_helper_iostream*>(this) =
            memory_helper_iostream(newData, newData+newCapacity);

        seek(offset);
        truncate(originalSize);
        return true;
    }

private:
    Allocator mAllocator;
};

typedef memory_iostream<> MemoryDuplexStream;
typedef memory_helper_iostream ExistingMemoryBlock;

namespace {
    template<typename T>
    T clamp(T value, T min, T max) {
        if(value < min) {
            return min;
        } else if(value > max) {
            return max;
        }
        return value;
    }
}
template<typename Allocator=std::allocator<uint8_t>>
class heap_iostream {
public:
    template<typename Allocator2>
    heap_iostream(const Allocator2 &allocator)
        : mAllocator(std::forward(allocator)) {

    }
    heap_iostream() {}


    ~heap_iostream() {
        clear();
    }

    heap_iostream &operator=(heap_iostream &&other) {
        mList = std::move(other.mList);
        mAllocator = std::move(other.mAllocator);
        mCurIndex = other.mCurIndex;
        mCurStream = other.mCurStream;
        mMinBlockSize = other.mMinBlockSize;
        mMaxBlockSize = other.mMaxBlockSize;

        other.mCurIndex     = 0;
        other.mCurStream    = nullptr;
        return *this;
    }
    heap_iostream(const heap_iostream&)             =default;
    heap_iostream(heap_iostream&& other) { *this = std::move(other); }



    /** ensures atleast `size` bytes of data is already preallocated. */
    void reserve(std::streamsize size) {
        std::streamsize capacity = this->capacity();
        if(capacity > size) {
            return;
        }

        std::streamsize extra = size - capacity;
        add_buffer(extra);
    }

    void set_min_block_size(std::streamsize min) {
        mMinBlockSize = min;
        if(mMinBlockSize > mMaxBlockSize) {
            mMaxBlockSize = mMinBlockSize;
        }
    }

    void set_max_block_size(std::streamsize max) {
        mMaxBlockSize = max;
        if(mMaxBlockSize < mMinBlockSize) {
            mMinBlockSize = mMaxBlockSize;
        }
    }

    std::streamsize capacity() {
        std::streamsize totalSize = 0;
        for(memory_helper_iostream &stream : mList) {
            totalSize += stream.capacity();
        }
        return totalSize;
    }

    std::streamsize size() {
        std::streamsize totalSize = 0;
        for(memory_helper_iostream &stream : mList) {
            totalSize += stream.size();
        }
        return totalSize;
    }

    void clear() {
        for(memory_helper_iostream &stream : mList) {
            mAllocator.deallocate(stream.data(), stream.capacity());
        }
        mList.clear();
    }

    std::streamsize write(const void *data, std::streamsize nBytes) {
        const_memory_helper_iostream chunk(data, nBytes);
        chunk.truncate(nBytes);

        while(!chunk.eof()) {
            std::streamsize remainder = chunk.size() - chunk.tell();
            if(mCurStream == nullptr || mCurStream->eof()) {
                next(remainder);
            }
            std::streamsize written = mCurStream->write(chunk.cursor(), remainder);
            chunk.seek(written, std::ios::cur);
        }
        return chunk.tell();
    }

    std::streamsize read(void *data, std::streamsize nBytes) {
        if(mList.size() == 0){
            return 0;
        }
        memory_helper_iostream chunk(data, nBytes);

        while(!chunk.eof()) {

            std::streamsize remainder = chunk.capacity() - chunk.tell();
            std::streamsize readBytes = mCurStream->read(chunk.cursor(), remainder);
            chunk.write(readBytes);
            if(mCurStream->eof()) {
                if(!next_read()) {
                    break;
                }
            }
        }
        return chunk.size();
    }

    std::streamoff tell() const {
        if(mList.size() == 0) {
            return 0;
        }
        std::streamoff offset = 0;
        for(std::size_t i = 0; i < mCurIndex; ++i) {
            offset += mList[i].size();
        }
        offset += mCurStream->size();
        return offset;
    }

    std::streamoff seek(std::streamoff offset, std::ios_base::seekdir dir = std::ios_base::beg) {
        switch(dir) {
        default:
        case std::ios_base::beg: break;
        case std::ios_base::cur: offset += tell(); break;
        case std::ios_base::end: offset += size(); break;
        }

        std::streamoff curOff = 0;
        for(std::size_t i = 0; i < mList.size(); ++i) {
            memory_helper_iostream &stream = mList[i];
            if(stream.size() == 0) {
                mCurIndex = i;
                mCurStream = &stream;
                break;
            }
            std::streamoff endOff = stream.size() + curOff;
            if(offset < endOff) {
                mCurIndex = i;
                mCurStream = &stream;
                mCurStream->seek(offset - curOff);
                curOff = offset;
                break;
            }
        }

        return curOff;
    }
private:
    void add_buffer(std::streamsize size) {
        std::streamsize roundTo = sizeof(uint8_t*)*2;
        while((size%roundTo) != 0) {
            ++size;
        }
        size = clamp(size, mMinBlockSize, mMaxBlockSize);
        mList.push_back(
            memory_helper_iostream(mAllocator.allocate(size), size)
        );
        mList.back().truncate(0);
        // adding to vector invalidates so lets get back our stream
        mCurStream = &mList[mCurIndex];
    }

    void next(std::streamsize size) {
        ++mCurIndex;
        if(mCurStream == nullptr) {
            mCurIndex = 0;
        }

        if(mCurIndex == mList.size()) {
            add_buffer(size);
        }

        mCurStream = &mList[mCurIndex];
        mCurStream->seek(0);
    }

    bool next_read() {
        if(mCurIndex == mList.size()-1) {
            return false;
        }
        if(mList[mCurIndex+1].size() == 0) {
            return false;
        }
        ++mCurIndex;
        mCurStream = &mList[mCurIndex];
        mCurStream->seek(0);
        return true;
    }
    Allocator mAllocator;
    std::vector<memory_helper_iostream, Allocator> mList;
    std::size_t mCurIndex = 0;
    std::streamsize mMinBlockSize = 1024;
    std::streamsize mMaxBlockSize = 1024*8;
    memory_helper_iostream *mCurStream = nullptr;
};
