#pragma once

#include <vector>
#include <memory>

/* Naming things is so hard X_X

    name                                Chars
    pmr_istream                         22
    polymorphic_istream_base            24
    polymorphic_istream                 19
    pmr::polymorphic_istream_base       29
    std::pmr::polymorphic_istream_base  34
    poly_istream_base                   17
    istream_base                        12
    pmr_istream_base                    16
    pmr::istream_base                   17
    pmr::pmr_istream_base               21
    pmr_istream                         11
    std::pmr::pmr_istream               21

    pmr_istream it is! It's descriptive. Not too long as I forsee some people
    using it's name alot. And clear as c++17 added pmr = polymorphic namespace
    so people are already getting used to pmr=polymorphic namespace.

    omg we need new name. pmr stands for Polymorphic Memory Resource and not
    PolyMoRphic.
*/
class pmr_istream {
public:
    pmr_istream(){}
    // Because it's virtual, we can't guarantee a safe copy or move.
    // in which case we disable it all.
    pmr_istream              (pmr_istream&&)      =delete;
    pmr_istream &operator=   (pmr_istream&&)      =delete;
    pmr_istream              (const pmr_istream&) =delete;
    pmr_istream &operator=   (const pmr_istream&) =delete;

    virtual ~pmr_istream(){};
    virtual std::streamsize read(void *data, std::streamsize sizeBytes)=0;
};

/*
    Next hard one to name X_X. It's just a thing to wrap an existing istream
    to make it easy to make a pmr_istream out of it. It's ok for this one
    to be longer as it will be less used than pmr_istream.

    example:
        pmr_istream_wrapper<Stream> *stream = new pmr_istream_wrapper<Stream>(Stream(std::move(prev)));
        pmr_istream_sub<Stream> *stream = new pmr_istream_sub<Stream>(Stream(std::move(prev)));


    name                                chars
    12345678                            8
    pmr_istream_sub                  18
    pmr_istream_wrapper                 19
    std::pmr::pmr_istream_wrapper       29
    make_pmr_istream                    16
    make_shared                         11 # for comparison :D
    make_istream                        12
    basic_pmr_istream                   17 # doesn't make sense
    pmr_istream_facade                  18 # has negative emotion in english
    pmr_istream_wrap                    16
    pmr_istream_mask                    16
    to_pmr_istream                      14
    pmr_istream_subclass                20
    pmr_istream_sub                     15
    new_pmr_istream                     15

    pmr_istream_sub wins. We can have a to_pmr_istream function too but doesn't
    makes sense as it will return a pointer. maybe a new_pmr_istream so it's
    easier to incorporate an allocator too.
*/
template<typename T>
class pmr_istream_sub : public pmr_istream {
public:
    pmr_istream_sub(T &&stream) : mStream(std::move(stream)) {}
    std::streamsize read(void* data, std::streamsize sizeBytes) override {
        return mStream.read(data, sizeBytes);
    }

    /*  These should not be virtual as it would have problems because it's
        templated.
    */
    T &get_istream() {
        return mStream;
    }

    void set_istream(T &&stream) {
        mStream = std::move(stream);
    }
private:
    T mStream;
};

class pmr_ostream {
public:
    pmr_ostream(){}
    // Because it's virtual, we can't guarantee a safe copy or move.
    // in which case we disable it all.
    pmr_ostream              (pmr_ostream&&)      =delete;
    pmr_ostream &operator=   (pmr_ostream&&)      =delete;
    pmr_ostream              (const pmr_ostream&) =delete;
    pmr_ostream &operator=   (const pmr_ostream&) =delete;

    virtual ~pmr_ostream(){};
    virtual std::streamsize write(const void *data, std::streamsize sizeBytes)=0;
};

template<typename T>
class pmr_ostream_sub : public pmr_ostream {
public:
    pmr_ostream_sub(T &&stream) : mStream(std::move(stream)) {}
    std::streamsize write(const void* data, std::streamsize sizeBytes) override {
        return mStream.read(data, sizeBytes);
    }

    /*  These should not be virtual as it would have problems because it's
        templated.
    */
    T &get_ostream() {
        return mStream;
    }

    void set_ostream(T &&stream) {
        mStream = std::move(stream);
    }
private:
    T mStream;
};

/** Holds a reference to the stream. This allows for a stream to not own
    a base stream. For example when linking up in a filter.
*/
template<typename T>
class ref_istream {
public:
    ref_istream(T &stream): mStream(&stream) {
    }

    std::streamsize read(void* data, std::streamsize nBytes) {
        return mStream->read(data, nBytes);
    }


    void link(T &stream) {
        mStream = &stream;
    }
private:
    T *mStream;
};

/** Holds a reference to the stream. This allows for a stream to not own
    a base stream. For example when linking up in a filter.
*/
template<typename T>
class ref_ostream {
public:
    ref_ostream(T &stream): mStream(&stream) {
    }

    std::streamsize write(void* data, std::streamsize nBytes) {
        return mStream->write(data, nBytes);
    }


    void link(T &stream) {
        mStream = &stream;
    }
private:
    T *mStream;
};


/* This class is meant for flexibility over runtime performance */
class chain_istream {
public:
    /* @todo: this needs a better name */
    typedef ref_istream<pmr_istream> ChainStream;

    template<typename T>
    chain_istream(T &&input) {
        ChainItem item;
        item.stream.reset(new pmr_istream_sub<T>(std::move(input)));
        mChain.push_back(std::move(item));
    }

    template<typename Stream>
    Stream &push_new() {
        ChainItem item;
        ChainStream prev(*(mChain.back().stream));
        pmr_istream_sub<Stream> *stream = new pmr_istream_sub<Stream>(Stream(std::move(prev)));
        item.stream.reset(stream);
        mChain.push_back(std::move(item));
        return stream->get_istream();
    }

    template<typename Stream, typename... Args>
    Stream &push_new(Args... args) {
        ChainItem item;
        ChainStream prev(*(mChain.back().stream));
        pmr_istream_sub<Stream> *stream = new pmr_istream_sub<Stream>(Stream(std::move(prev), args...));
        item.stream.reset(stream);
        mChain.push_back(std::move(item));
        return stream->get_istream();
    }

    template<typename Stream>
    void push(Stream &&stream) {
        ChainStream prev;
        prev.link(*(mChain.back().stream));
        stream.set_istream(prev);

        ChainItem item;
        item.stream = std::unique_ptr<pmr_istream_sub<Stream>>(std::move(stream));
        mChain.push_back(std::move(item));
    }
    std::streamsize read(void *data, std::streamsize nBytes) {
        return mChain.back().stream->read(data, nBytes);
    }
private:
    class ChainItem {
    public:
        std::unique_ptr<pmr_istream> stream;
    };
    std::vector<ChainItem> mChain;
};


// this class is useless
template<typename T>
class InputStreamHolder {
public:
    InputStreamHolder(){}
    explicit InputStreamHolder(T &&stream) {
        mStream = std::move(stream);
    }

    InputStreamHolder(InputStreamHolder&& other) {
        (*this) = std::move(other);
    }
    InputStreamHolder &operator=(InputStreamHolder&&other) {
        mStream = std::move(other.mStream);
    }
    InputStreamHolder(const InputStreamHolder&)=delete;
    InputStreamHolder &operator=(const InputStreamHolder&)=delete;
    InputStreamHolder &operator=(T&& stream) {
        mStream = std::move(stream);
    }
    std::streamsize read(void *data, std::streamsize nBytes) {
        return mStream.read(data, nBytes);
    }
    // just to make it easier to access specific function of what it's holding
    T *operator->() { return &mStream; }
    T &get() { return mStream; }

private:
    // perhaps optional is better here because it can be invalid.
    // If no optional then T must be able to hold nothing.
    T mStream;
};

// @todo do the std::if_enabled thing to only allow pod types for this
template<typename T, typename POD>
InputStreamHolder<T> &operator >>(InputStreamHolder<T> &stream, POD &output) {
    stream.read(&output, sizeof(POD));
}


